@@ -4,6 +4,8 @@
 from __future__ import unicode_literals
 from __future__ import print_function
 
+import copy
+
 from ldapom.cdef import libldap, ffi
 from ldapom import attribute
 from ldapom import compat
@@ -165,7 +167,8 @@ class LDAPConnection(object):
         if name in self._attribute_types_by_name:
             return self._attribute_types_by_name[name]
         else:
-            raise error.LDAPAttributeNameNotFoundError
+            raise error.LDAPAttributeNameNotFoundError(
+                    'Attribute type "{}" not found.'.format(name))
 
     @_retry_reconnect
     def can_bind(self, bind_dn, bind_password):
@@ -268,15 +271,15 @@ class LDAPConnection(object):
         """Search without retry_reconnect."""
         try:
             for dn, attributes_dict in self._raw_search(*args, **kwargs):
-                entry = LDAPEntry(self, compat._decode_utf8(dn),
-                        attributes=set())
+                entry = LDAPEntry(self, compat._decode_utf8(dn))
+                entry._attributes = set()
                 for name, value in attributes_dict.items():
                     # TODO: Create the right type of LDAPAttribute here
                     attribute_type = self.get_attribute_type(
                             compat._decode_utf8(name))
                     attribute = attribute_type(compat._decode_utf8(name))
                     attribute._set_ldap_values(value)
-                    entry.attributes.add(attribute)
+                    entry._attributes.add(attribute)
                 yield entry
         except error.LDAPNoSuchObjectError:
             # If the search returned without results, "return" an empty generator.
@@ -349,31 +352,30 @@ class LDAPConnection(object):
         """
         entry_exists = entry.exists()
         # Refuse to save if attributes have not been fetched or set explicitly.
-        if entry.attributes is None:
+        if entry._attributes is None:
             raise error.LDAPomError("Cannot save without attributes "
                     "previously fetched or set.")
 
-        # Temporary attribute set that will contain deleted attributes as
-        # LDAPAttribute objects without any values.
-        save_attributes = entry.attributes.copy()
-        # Don't try to save empty attributes as this fails if the entry does
-        # not exist on the server yet.
-        if not entry_exists:
-            save_attributes = set(filter(lambda attr: len(attr._values) > 0,
-                    save_attributes))
-
-        deleted_attribute_names = entry._old_attribute_names.difference(
-                [a.name for a in entry.attributes])
-        for name in deleted_attribute_names:
-            attribute_type = self.get_attribute_type(name)
-            save_attributes.add(attribute_type(name))
+        if entry_exists:
+            assert entry._fetched_attributes is not None
+            changed_attributes = entry._attributes - entry._fetched_attributes
+            # Deleted attributes are represented as empty attributes to the LDAP server.
+            deleted_attribute_names = (frozenset(a.name for a in entry._fetched_attributes)
+                    - frozenset(a.name for a in entry._attributes))
+            for deleted_name in deleted_attribute_names:
+                deleted_attribute_type = self.get_attribute_type(deleted_name)
+                changed_attributes.add(deleted_attribute_type(deleted_name))
+        else:
+            # Don't try to save empty attributes as this fails if the entry does
+            # not exist on the server yet.
+            changed_attributes = set(filter(lambda attr: len(attr._values) > 0, entry._attributes))
 
         # Keep around references to pointers to owned memory with data that is
         # still needed.
         prevent_garbage_collection = []
 
-        mods = ffi.new("LDAPMod*[{}]".format(len(save_attributes) + 1))
-        for i, attribute in enumerate(save_attributes):
+        mods = ffi.new("LDAPMod*[{}]".format(len(changed_attributes) + 1))
+        for i, attribute in enumerate(changed_attributes):
             mod = ffi.new("LDAPMod *")
             prevent_garbage_collection.append(mod)
 
@@ -393,7 +395,7 @@ class LDAPConnection(object):
             mod.mod_vals = {"modv_strvals": modv_strvals}
 
             mods[i] = mod
-        mods[len(save_attributes)] = ffi.NULL
+        mods[len(changed_attributes)] = ffi.NULL
 
         if entry_exists:
             err = libldap.ldap_modify_ext_s(self._ld,
@@ -407,7 +409,7 @@ class LDAPConnection(object):
                     ffi.NULL, ffi.NULL)
         handle_ldap_error(err)
 
-        entry._old_attribute_names = set([a.name for a in entry.attributes])
+        entry._fetched_attributes = copy.deepcopy(entry._attributes)
 
     @_retry_reconnect
     def fetch(self, entry):
@@ -419,8 +421,8 @@ class LDAPConnection(object):
         try:
             fetched_entry = next(self._search(base=entry.dn,
                 scope=libldap.LDAP_SCOPE_BASE))
-            entry.attributes = fetched_entry.attributes
-            entry._old_attribute_names = set([a.name for a in entry.attributes])
+            entry._attributes = fetched_entry._attributes
+            entry._fetched_attributes = copy.deepcopy(entry._attributes)
         except StopIteration:
             raise error.LDAPNoSuchObjectError()
 
